package teralg;

// Bináris fa csomópont struktúra.
// Privát láthatósággal a BinarySearchTree osztályba is beágyazható, és elrejthető kifelé.
class Node {
    public Node parent;
    public Node left;
    public Node right;
    public int key;
    
    public Node(int k) {
        key = k;
    }
}

// Bináris keresőfa típus.
public class BinarySearchTree {
    // Csak a fa gyökér csomópontját szükséges tárolni, a többi elérhető rajta keresztül.
    private Node root = null;
    
    // InOrder bejárás kiíratással.
    public void inorder() {
        inorder(root);
        System.out.println();
    }
    
    // InOrder bejárás (paraméterezhető).
    private void inorder(Node n) {
        if(n != null) {
            inorder(n.left);
            System.out.print(n.key + " ");
            inorder(n.right);
        }
    }
    
    // Kulcs keresése.
    // Visszatérési érték: igaz, ha tartalmazza a fa a kulcsot; egyébként hamis.
    public boolean find(int key) {
        Node p = root;
        while(p != null && p.key != key) {
            if(key < p.key)
                p = p.left;
            else
                p = p.right;
        }
        return p != null;
    }
    
    // Kulcs beszúrása.
    // Visszatérési érték: igaz, ha sikerült beszúrni; hamis, ha már tartalmazta a fa a kulcsot.
    public boolean insert(int key) {
        // Először megkeressük a megfelelő szülő elemet a fában a beszúráshoz.
        Node p = root;
        Node q = null;
        while(p != null && p.key != key) {
            q = p;
            if(key < p.key)
                p = p.left;
            else
                p = p.right;
        }
        if(p != null)
            return false;
        
        // Beszúrjuk az új csomópontot q alá.
        Node n = new Node(key);
        n.parent = q;
        if(q == null)
            root = n;
        else {
            if(key < q.key)
                q.left = n;
            else
                q.right = n;
        }
        return true;
    }
    
    // Kulcs törlése.
    // Visszatérési érték: igaz, ha sikerült törölni; hamis, ha nem tartalmazta a fa a kulcsot.
    public boolean delete(int key) {
        // Először megkeressük a törlendő csomópontot.
        Node p = root;
        while(p != null && p.key != key) {
            if(key < p.key)
                p = p.left;
            else
                p = p.right;
        }
        if(p == null)
            return false;
        
        // p: törlendő csomópont.
        // q: törlendő csomópont szülője.
        Node q = p.parent;
        // 1. eset: ha nincsen gyereke a törlendő csomópontnak.
        if(p.left == null && p.right == null) {
            // Ha a gyökér elemet töröljük, akkor üres fát kapunk.
            if(q == null)
                root = null;
            // Egyébként kiláncoljuk p-t q alól.
            else {
                if(p.key < q.key)
                    q.left = null;
                else
                    q.right = null;
            }
            // Bizotsítjuk, hogy a törölt p-ben se maradjon érvénytelen kivezető hivatkozás.
            p.parent = null;
        }
        // 2. eset: ha csak bal gyereke van a törlendő csomópontnak.
        else if(p.left != null && p.right == null) {
            // r: törlendő csomópont egyetlen gyereke.
            Node r = p.left;
            r.parent = q;
            
            // Ha a gyökér elemet töröljük, akkor annak gyereke lesz az új gyökér.
            if(q == null)
                root = r;
            // Egyébként átláncoljuk r-t q alá.
            else {
                if(p.key < q.key)
                    q.left = r;
                else
                    q.right = r;
            }
            // Bizotsítjuk, hogy a törölt p-ben se maradjon érvénytelen kivezető hivatkozás.
            p.parent = p.left = null;
        }
        // 3. eset: ha csak jobb gyereke van a törlendő csomópontnak.
        else if(p.left == null && p.right == null) {
            // r: törlendő csomópont egyetlen gyereke.
            Node r = p.right;
            r.parent = q;
            
            // Ha a gyökér elemet töröljük, akkor annak gyereke lesz az új gyökér.
            if(q == null)
                root = r;
            // Egyébként átláncoljuk r-t q alá.
            else {
                if(p.key < q.key)
                    q.left = r;
                else
                    q.right = r;
            }
            // Bizotsítjuk, hogy a törölt p-ben se maradjon érvénytelen kivezető hivatkozás.
            p.parent = p.right = null;
        }
        // 4. eset: ha mindkét oldali gyereke van a csomópontnak.
        else {
            // r: p jobb oldali részfájának bal alsó (legkisebb) eleme.
            Node r = p.right;
            while(r.left != null)
                r = r.left;
            
            // Töröljük r-t (garantáltan nem lehet mindkét oldali gyereke).
            delete(r.key);
            // r kulcsát átírjuk p-be.
            p.key = r.key;
        }
        return true;
    }
}
